--[[
  Copyright (c) 2021 by Plexim GmbH
  All rights reserved.

  A free license is granted to anyone to use this software for any legal
  non safety-critical purpose, including commercial applications, provided
  that:
  1) IT IS NOT USED TO DIRECTLY OR INDIRECTLY COMPETE WITH PLEXIM, and
  2) THIS COPYRIGHT NOTICE IS PRESERVED in its entirety.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
--]] --
local U = {}

math.randomseed(os.time())
local random = math.random

local LogText = StringList:new()

function U.log(line)
  LogText:append(line)
end

function U.dumpLog(filename)
  local file, e = io.open(filename, "w")
  if file == nil then
    return e
  end
  io.output(file)
  for _, v in ipairs(LogText) do
    io.write(v);
  end
  file.close()
end

function U.fileOrDirectoryExists(file)
  return (Plecs:FileExists(file) or Plecs:DirectoryExists(file))
end

function U.copyTemplateFile(src, dest, subs)
  local file, e = io.open(src, "rb")
  if file == nil then
    return e
  end
  local src_content = file:read("*all")
  io.close(file)
  local dest_content

  file = io.open(dest, "rb")
  if (file == nil) then
    dest_content = nil
  else
    dest_content = file:read("*all")
    io.close(file)
    dest_content = string.gsub(dest_content, "\r", "")
  end

  if subs ~= nil then
    for _, v in pairs(subs) do
      local before = v["before"]
      local after = v["after"]
      src_content = string.gsub(src_content, before, after)
    end
  end

  src_content = string.gsub(src_content, "\r", "")

  if not (src_content == dest_content) then
    file, e = io.open(dest, "w")
    if file == nil then
      return e
    end
    io.output(file)
    io.write(src_content)
    file.close()
  end
end

function U.getFromArrayOrScalar(field, index, majordim)
  if #field == 1 then
    return field[1]
  elseif #field == majordim then
    return field[index]
  else
    return nil
  end
end

function U.round(num, numDecimalPlaces)
  if numDecimalPlaces and numDecimalPlaces > 0 then
    local mult = 10 ^ numDecimalPlaces
    return math.floor(num * mult + 0.5) / mult
  end
  return math.floor(num + 0.5)
end

function U.guid()
  local template = '{0xXX, 0xXX, 0xXX, 0xXX, 0xXX, 0xXX, 0xXX, 0xXX}'
  return string.gsub(template, '[X]', function(c)
    local v = (c == 'X') and random(0, 0xf)
    return string.format('%x', v)
  end)
end

function U.isValidCName(name)
  if string.match(name, '[^a-zA-Z0-9_]') ~= nil then
    return false
  end
  if string.match(string.sub(name, 1, 1), '[0-9]') ~= nil then
    return false
  end
  return true
end

function U.isPositiveScalar(val)
  if type(val) ~= 'number' then
    return false
  end
  if val < 1 then
    return false
  end
  return true
end

function U.isPositiveIntScalar(val)
  if type(val) ~= 'number' then
    return false
  end
  if math.floor(val) ~= val or val < 1 then
    return false
  end
  return true
end

function U.isNonNegativeIntScalar(val)
  if type(val) ~= 'number' then
    return false
  end
  if math.floor(val) ~= val or val < 0 then
    return false
  end
  return true
end

function U.isPositiveIntScalarOrArray(val)
  if type(val) == 'table' then
    for _, v in pairs(val) do
      if not U.isNonNegativeIntScalar(v) then
        return false
      end
    end
    return true
  else
    return U.isNonNegativeIntScalar(val)
  end
end

function U.getInstallDir()
  local installDir = Target.Variables.BUILD_ROOT

  if Target.Variables.genOnly == 1 then
    installDir = Target.Variables.installDir:gsub('%"+', ''):gsub('\\+', '/') -- remove quotes, make all forward slashes
    local tempDir = Target.Variables.BUILD_ROOT:gsub('%"+', ''):gsub('\\+', '/')
    if installDir .. Target.Variables.BASE_NAME .. "_codegen" == tempDir then -- when installDir entry is blank, returns current path
      installDir = Target.Variables.BUILD_ROOT --  use default codegen dir if installDir is blank
    end
  end
  return installDir
end

function U.copyTemplateToBuildDir(template, destname, dict)
  local installDir = U.getInstallDir()
  if not U.fileOrDirectoryExists(installDir .. "/") then
    return "The directory '%s' does not exist." % {installDir}
  end
  U.copyTemplateFile(template, installDir .. "/" .. destname, dict)
end

return U
